/* jld.c - load and view JPEG file
   Copyright (C) 1993 NIIMI Satoshi

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include "jinclude.h"
#include "getopt.h"
#include <stdlib.h>
#include <conio.h>
#include <setjmp.h>
#include <sys/dos.h>
#include <sys/iocs.h>

#define getgramaddr(y) ((unsigned short *) (0xc00000 | (((y) & 0x1ff) << 10)))

#undef alloca
#define alloca __builtin_alloca

extern void setSquarePixelMode (void);

static char rcsid[] = "$Id: jld.c,v 1.7 1993/07/03 02:03:09 sa2c Exp $";

extern unsigned char dither_table[4][4][256];

unsigned short **gram_buffer;

static int row_offset;

static int mode = 'h';
static int verbose_level = 0;
static int no_scroll = FALSE;
static int no_key = FALSE;
static int gray_output = FALSE;

static char *progname;

static jmp_buf setjmp_buffer;

void
screen_init (void)
{
  static int initialized = 0;

  if (!initialized)
    {
      _dos_c_fnkmod (2);
      _dos_c_curoff ();
      if (mode == 'h')
	setSquarePixelMode ();
      else
	_dos_c_width (5);
      initialized = 1;
    }
}

inline static void
scroll (int x, int y)
{
  _iocs_scroll (0, x, y);
  _iocs_scroll (1, x, y);
  _iocs_scroll (2, x, y);
  _iocs_scroll (3, x, y);
}

void *
xmalloc (size_t size)
{
  void *p = malloc (size);
  if (p == NULL)
    {
      fprintf (stderr, "virtual memory exhausted\n");
      exit (2);
    }
  return p;
}

void
output_init (decompress_info_ptr cinfo)
{
  unsigned short *p;
  int i;

  row_offset = 0;
  _iocs_wipe ();
  scroll (0, 0);
  if (no_scroll)
    cinfo->image_height = 512;
  gram_buffer = xmalloc (sizeof (*gram_buffer) * cinfo->image_height);
  p = xmalloc (sizeof (*p) * cinfo->image_width * cinfo->image_height);
  for (i = 0; i < cinfo->image_height; i++)
    gram_buffer[i] = &p[i * cinfo->image_width];
}

void
put_color_map (decompress_info_ptr cinfo, int num_colors, JSAMPARRAY colormap)
{
  fprintf (stderr, "put_color_map called: there's a bug here somewhere!\n");
}

inline static void
draw (short col, short row, short width, short height)
{
  short y;
  short width1, width2;
  int stack;
  unsigned short *gram;
  unsigned short *buf;

  width1 = 512 - (col & 511);
  if (width1 > width)
    {
      width1 = width;
      width2 = 0;
    }
  else
    width2 = width - width1;
  stack = _dos_super (0);
  for (y = row; y < row + height; y++)
    {
      short w1 = width1, w2 = width2;

      buf = &gram_buffer[y][col];
      gram = getgramaddr (y) + (col & 511);
      while (w1--)
	*gram++ = *buf++;
      if (w2)
	{
	  gram = getgramaddr (y);
	  while (w2--)
	    *gram++ = *buf++;
	}
    }
  _dos_super (stack);
}

inline static void
put_pixels_gray (unsigned short *gram, JSAMPROW ptr0, short width, short row)
{
  if (width < 0)
    return;
  while (width--)
    {
      unsigned char base = *ptr0++;
      unsigned char b1 = base >> 3;
      unsigned char b2 = (base >> 2) & 1;
      *gram++ = (b1 << 11) | (b1 << 6) | (b1 << 1) | b2;
    }
}

inline static void
put_pixels (unsigned short *gram, JSAMPROW ptr0, JSAMPROW ptr1, JSAMPROW ptr2,
	    short width, short row)
{
  unsigned char (*dither_base)[256] = dither_table[row & 3];
  short int col = 0;

  if (width < 0)
    return;
  while (width--)
    {
      unsigned char *dither = dither_base[col];
      unsigned char r = dither[GETJSAMPLE (*ptr0++)];
      unsigned char g = dither[GETJSAMPLE (*ptr1++)];
      unsigned char b = dither[GETJSAMPLE (*ptr2++)];
      *gram++ = (g << 11) | (r << 6) | (b << 1) | 1;
      if (++col > 3)
	col = 0;
    }
}

void
put_pixel_rows (decompress_info_ptr cinfo, int num_rows, JSAMPIMAGE pixel_data)
{
  short row;
  short width;

  if (_iocs_b_keysns () && _iocs_b_keyinp () & 0xff)
    longjmp (setjmp_buffer, 1);
  width = cinfo->image_width;

  for (row = 0; row < num_rows; row++)
    {
      if (cinfo->out_color_space == CS_GRAYSCALE)
	{
	  put_pixels_gray (gram_buffer[row_offset + row],
			   pixel_data[0][row],
			   width, row_offset + row);
	}
      else
	{
	  put_pixels (gram_buffer[row_offset + row],
		      pixel_data[0][row], pixel_data[1][row], pixel_data[2][row],
		      width, row_offset + row);
	}
    }
  if (row_offset + num_rows > 512)
    scroll (0, (row_offset + num_rows) & 511);
  draw (0, row_offset, min (512, width), num_rows);
  row_offset += num_rows;
  if (_iocs_b_keysns () && _iocs_b_keyinp () & 0xff)
    longjmp (setjmp_buffer, 1);
}

int pressed_key = 0;

void
scroll_view (decompress_info_ptr cinfo)
{
  int x, y;
  int c;
  int lastx, lasty;
  int maxx, maxy;

  x = 0;
  y = max (512, row_offset) - 512;

  maxx = max (512, cinfo->image_width) - 512;
  maxy = max (512, cinfo->image_height) - 512;

  putch ('\a');
  for (;;)
    {
      lastx = x;
      lasty = y;
      while ((c = _iocs_b_keyinp ()) == 0)
	;
      switch (c >> 8)
	{
	case 0x3b: case 0x23:
	  x -= 16;
	  if (x < 0)
	    x = 0;
	  break;
	case 0x3e: case 0x24:
	  y += 16;
	  if (y > maxy)
	    y = maxy;
	  break;
	case 0x3c: case 0x25:
	  y -= 16;
	  if (y < 0)
	    y = 0;
	  break;
	case 0x3d: case 0x26:
	  x += 16;
	  if (x > maxx)
	    x = maxx;
	  break;
	default:
	  if (c & 0xff)
	    {
	      pressed_key = c & 0xff;
	      return;
	    }
	}
      scroll (x & 511, y & 511);
      if (x < lastx)
	draw (x, y, lastx - x, min (512, cinfo->image_height));
      else if (x > lastx)
	draw (lastx + 512, y, x - lastx, min (512, cinfo->image_height));
      if (y < lasty)
	draw (x, y, min (512, cinfo->image_width), lasty - y);
      else if (y > lasty)
	draw (x, lasty + 512, min (512, cinfo->image_width), y - lasty);
    }
}

void
output_term (decompress_info_ptr cinfo)
{
  if (!no_key)
    scroll_view (cinfo);
  free (*gram_buffer);
  free (gram_buffer);
}

void
d_ui_method_selection (decompress_info_ptr cinfo)
{
  /* if grayscale input, force grayscale output; */
  /* else leave the output colorspace as set by main routine. */
  if (cinfo->jpeg_color_space == CS_GRAYSCALE)
    cinfo->out_color_space = CS_GRAYSCALE;

  /* select output routines */
  cinfo->methods->output_init = output_init;
  cinfo->methods->put_color_map = put_color_map;
  cinfo->methods->put_pixel_rows = put_pixel_rows;
  cinfo->methods->output_term = output_term;
}


int
read_JPEG_file (char *filename)
{
  struct Decompress_info_struct cinfo;
  struct Decompress_methods_struct dc_methods;
  struct External_methods_struct e_methods;

  if ((cinfo.input_file = fopen (filename, "rb")) == NULL)
    {
      fprintf (stderr, "%s: can't open %s\n", progname, filename);
      return 0;
    }

  screen_init ();

  cinfo.output_file = NULL;	/* No actual output file */

  /* Set up links to method structures. */
  cinfo.methods = &dc_methods;
  cinfo.emethods = &e_methods;

  if (setjmp (setjmp_buffer))
    {
      cinfo.image_height = row_offset;
      if (!no_key)
	scroll_view (&cinfo);
      free (*gram_buffer);
      free (gram_buffer);
      fclose (cinfo.input_file);
      return 1;
    }
  jselmemmgr (&e_methods);	/* select std memory allocation routines */
  jselerror (&e_methods);	/* select std error handling routines */
  dc_methods.d_ui_method_selection = d_ui_method_selection;
  j_d_defaults (&cinfo, TRUE);
  if (gray_output)
    cinfo.out_color_space = CS_GRAYSCALE;
  cinfo.emethods->trace_level = verbose_level;
  jselrjfif (&cinfo);
  jpeg_decompress (&cinfo);
  fclose (cinfo.input_file);
  return 1;			/* indicate success */
}

static int fnkmod;

void
cleanup (void)
{
  _dos_c_fnkmod (fnkmod);
  _dos_c_curon ();
}

void
print_usage (void)
{
  fprintf (stderr, "Usage: %s [-ghlskv] filename...\n", progname);
}

__volatile__ void
help (void)
{
  print_usage ();
  fprintf (stderr, "-g ... force gray-scale output\n");
  fprintf (stderr, "-h ... view 768x512 mode\n");
  fprintf (stderr, "-l ... view 512x512 mode (without column reducation)\n");
  fprintf (stderr, "-s ... no scroll (up to 512line)\n");
  fprintf (stderr, "-k ... don't wait key pression\n");
  fprintf (stderr, "-v ... verbose\n");
  exit (1);
}

__volatile__ void
usage (void)
{
  print_usage ();
  exit (1);
}

void
parse_opt (int argc, char **argv)
{
  int c;

  opterr = 0;

  while ((c = getopt (argc, argv, "ghlskv")) != -1)
    {
      switch (c)
	{
	case 'g':
	  gray_output = TRUE;
	  break;
	case 'h':
	  mode = 'h';
	  break;
	case 'l':
	  mode = 'l';
	  break;
	case 'v':
	  verbose_level++;
	  break;
	case 's':
	  no_scroll = TRUE;
	  break;
	case 'k':
	  no_key = TRUE;
	  break;
	case '?':
	  help ();
	  break;
	default:
	  fprintf (stderr, "Unknown option `-%c'.\n", optopt);
	  usage ();
	  break;
	}
    }
}

int
main (int argc, char **argv)
{
  int i;

  progname = argv[0];
  parse_opt (argc, argv);
  if (optind == argc)
    usage ();
  atexit (cleanup);
  fnkmod = _dos_c_fnkmod (-1);
  for (i = optind; i < argc; i++)
    {
      if (read_JPEG_file (argv[i]) == 0)
	exit (1);
      if (pressed_key == 'q' || pressed_key == 'Q')
	break;
      else if (pressed_key == '\x1b')
	exit (0);
    }
  _dos_c_width (0);
  exit (0);
}
